import { useMemo } from 'react';
import BigNumber from 'bignumber.js';
import { useOnomyApp } from '../../wallet/providers/OnomyApp';
/* eslint-disable class-methods-use-this */
import {
  Bar,
  ErrorCallback,
  HistoryCallback,
  IDatafeedChartApi,
  LibrarySymbolInfo,
  OnReadyCallback,
  PeriodParams,
  ResolutionString,
  ResolveCallback,
  SearchSymbolsCallback,
  SubscribeBarsCallback,
} from './types/tv-datafeed-api';
import { OnexHistoryFrames } from './OnexHistoryFrames';

const PERIOD_MAP = {
  '1': 'Minute',
  '15': 'QuarterHour',
  '60': 'Hour',
  '480': 'QuarterDay',
  '1D': 'Day',
  '1W': 'Week',
};

const ONEX_SYMBOL_INFO: LibrarySymbolInfo = {
  name: 'bNOM',
  full_name: 'bNOM',
  description: 'bNOM',
  exchange: 'ONEX',
  listed_exchange: 'ONEX',
  format: 'price',
  has_intraday: true,
  has_no_volume: false,
  has_empty_bars: true,
  data_status: 'streaming',
  original_currency_code: 'ETH',
  currency_code: 'ETH',
  // @ts-ignore
  supported_resolutions: Object.keys(PERIOD_MAP),
  intraday_multipliers: Object.keys(PERIOD_MAP),
  ticker: 'bNOM',
  timezone: 'Etc/UTC',
  type: 'crypto',
  session: '24x7',
  minmov: 1,
  pricescale: 100000,
};

type BarSubscription = {
  symbolInfo: LibrarySymbolInfo;
  resolution: ResolutionString;
  onTick: SubscribeBarsCallback;
  listenerGuid: string;
  onResetCacheNeededCallback: () => void;
};

function serializeBar(bar: Bar) {
  return [bar.close, bar.open, bar.high, bar.low, bar.volume, bar.time].join('-');
}

// @ts-ignore
export class OnexTradingViewData implements IDatafeedChartApi {
  private onexHistoryFrames: OnexHistoryFrames;

  private barSubs: Record<string, BarSubscription>;

  private lastTick: Record<string, Bar>;

  private baseDenom: string;

  private quoteDenom: string;

  private timer: any;

  constructor(onexHistoryFrames: OnexHistoryFrames, baseDenom: string, quoteDenom: string) {
    this.onexHistoryFrames = onexHistoryFrames;
    this.barSubs = {};
    this.lastTick = {};
    this.baseDenom = baseDenom;
    this.quoteDenom = quoteDenom;
    this.onBlockMined = this.onBlockMined.bind(this);

    this.timer = setInterval(this.onBlockMined.bind(this), 1000) as any;
  }

  private async onBlockMined() {
    const subs = Object.values(this.barSubs);

    // eslint-disable-next-line no-restricted-syntax
    for (const { symbolInfo, resolution, onTick } of subs) {
      this.getBars(
        symbolInfo,
        resolution,
        { countBack: 1, from: 0, to: 0, firstDataRequest: false },
        res => {
          const last = res.slice().pop();
          if (last) {
            const lastTick = this.lastTick[resolution];
            if (!lastTick || serializeBar(last) !== serializeBar(lastTick)) {
              this.lastTick[resolution] = last;
              onTick(last);
            }
          }
        },
        // eslint-disable-next-line no-console
        err => console.error(err)
      );
    }
  }

  public destroy() {
    clearInterval(this.timer);
  }

  public onReady(cb: OnReadyCallback) {
    // TradingView requires this setTimeout
    setTimeout(
      () =>
        cb({
          exchanges: [
            {
              value: ONEX_SYMBOL_INFO.exchange,
              name: ONEX_SYMBOL_INFO.exchange,
              desc: ONEX_SYMBOL_INFO.exchange,
            },
          ],
          symbols_types: [
            {
              value: ONEX_SYMBOL_INFO.type,
              name: ONEX_SYMBOL_INFO.type,
            },
          ],
          supported_resolutions: ONEX_SYMBOL_INFO.supported_resolutions,
          currency_codes: ['ETH'],
          units: {},
          supports_marks: false,
          supports_timescale_marks: false,
          supports_time: true,
        }),
      0
    );
  }

  public async searchSymbols(
    userInput: string,
    exchange: string,
    symbolType: string,
    onResult: SearchSymbolsCallback
  ) {
    onResult([
      {
        symbol: ONEX_SYMBOL_INFO.ticker!,
        full_name: ONEX_SYMBOL_INFO.name,
        description: ONEX_SYMBOL_INFO.description,
        exchange: ONEX_SYMBOL_INFO.exchange,
        ticker: ONEX_SYMBOL_INFO.ticker!,
        type: ONEX_SYMBOL_INFO.type,
      },
    ]);
  }

  resolveSymbol(symbolName: string, onSuccess: ResolveCallback) {
    // TradingView requires this setTimeout
    setTimeout(() => onSuccess(ONEX_SYMBOL_INFO), 0);
  }

  public async getBars(
    symbolInfo: LibrarySymbolInfo,
    resolution: string,
    periodParams: PeriodParams,
    onResult: HistoryCallback,
    onError: ErrorCallback
  ) {
    // console.log('periodParams', periodParams);

    try {
      const params = {
        // @ts-ignore
        frameType: PERIOD_MAP[resolution] || 'Hour',
        skipped: 0,
        baseDenom: this.baseDenom,
        quoteDenom: this.quoteDenom,
        from: periodParams.from,
        to: periodParams.to ?? undefined,
        first: Math.min(1000, periodParams.countBack),
      };

      const frames = await this.onexHistoryFrames.forceGet(params);
      const bars = frames
        .slice()
        .reverse()
        .map(
          (frame: {
            startTime: any;
            endTime: any;
            startPrice: any;
            maxPrice: any;
            minPrice: any;
            endPrice: any;
            volume: any;
          }) => {
            const time = parseInt(frame.startTime, 10) * 1000;
            return {
              time,
              open: frame.startPrice,
              high: frame.maxPrice,
              low: frame.minPrice,
              close: frame.endPrice,
              // TODO: do not hardcode decimal places
              volume: new BigNumber(frame.volume).shiftedBy(-18).toNumber(),
            };
          }
        );
      if (bars.length > 0) {
        onResult(bars, { noData: false });
      } else {
        onResult([], { noData: true });
      }

      bars.forEach(bar => {
        if ((this.lastTick[resolution]?.time ?? 0) <= bar.time) {
          this.lastTick[resolution] = bar;
        }
      });
    } catch (e) {
      // eslint-disable-next-line no-console
      console.error(e);
      onError(`${e}`);
    }
  }

  public subscribeBars(
    symbolInfo: LibrarySymbolInfo,
    resolution: ResolutionString,
    onTick: SubscribeBarsCallback,
    listenerGuid: string,
    onResetCacheNeededCallback: () => void
  ) {
    this.barSubs[listenerGuid] = {
      symbolInfo,
      resolution,
      onTick,
      listenerGuid,
      onResetCacheNeededCallback,
    };
  }

  public unsubscribeBars(listenerGuid: string) {
    delete this.barSubs[listenerGuid];
  }
}

export function useOnexTradingViewData(baseDenom: string, quoteDenom: string) {
  const { onexHistoryFrames } = useOnomyApp();

  try {
    return useMemo(() => {
      if (!onexHistoryFrames) return null;
      return new OnexTradingViewData(onexHistoryFrames, baseDenom, quoteDenom);
    }, [onexHistoryFrames, baseDenom, quoteDenom]);
  } catch (e) {
    console.warn('useOnexTradeData error', e);
    // Occurs in testing
    return null;
  }
}
