import { useCallback, useEffect, useMemo, useState } from 'react';
import BigNumber from 'bignumber.js';
import { OnexClient } from '@onomy/client';
import { DetailedFungibleAssetOnChain } from '@onomy/hub';
import { useAsyncPoll, useAsyncProcess, useAsyncStatuses } from '@onomy/react-utils';
import { useAssetBalanceUnlocked, useDetailedBalance, useFixedAmountInput } from '@onomy/react-hub';
import { Pool } from '@onomy/telescope/dist/codegen/market/pool';
import { useOnexDetailedPoolPrice, useOnexClient } from '../hooks';

const DEFAULT_AMOUNT_INPUTS_STATE = {
  amountBaseInputFixed: '0',
  amountQuoteInputFixed: '0',
  amountLastChanged: 'base' as 'base' | 'quote',
};

export function useAddDropsForm({
  baseAsset,
  quoteAsset,
  pool,
}: {
  baseAsset: DetailedFungibleAssetOnChain;
  quoteAsset: DetailedFungibleAssetOnChain;
  pool: Pool | null;
}) {
  const onex = useOnexClient();
  const [attemptTx, txStatus, resetTxStatus] = useAsyncProcess();
  const [txResult, setTxResult] = useState<null | Awaited<ReturnType<OnexClient['createDrop']>>>(
    null
  );

  const [baseBalance, baseBalanceStatus] = useAssetBalanceUnlocked(baseAsset);
  const [quoteBalance, quoteBalanceStatus] = useAssetBalanceUnlocked(quoteAsset);
  const balanceStatus = useAsyncStatuses([baseBalanceStatus, quoteBalanceStatus]);

  const [currentPricePerBase, currentPricePerBaseStatus] = useOnexDetailedPoolPrice({
    baseAsset,
    quoteAsset,
  });
  const [currentPricePerQuote, currentPricePerQuoteStatus] = useOnexDetailedPoolPrice({
    baseAsset: quoteAsset,
    quoteAsset: baseAsset,
  });
  const pair = useMemo(
    () => [baseAsset.id, quoteAsset.id].sort().join(','),
    [baseAsset.id, quoteAsset.id]
  );

  const [{ amountBaseInputFixed, amountQuoteInputFixed, amountLastChanged }, setInputState] =
    useState(DEFAULT_AMOUNT_INPUTS_STATE);

  const setAmountBaseFixed = useCallback((newAmountFixed: string) => {
    setInputState(state => ({
      ...state,
      amountBaseInputFixed: newAmountFixed,
      amountLastChanged: 'base',
    }));
  }, []);

  const setAmountQuoteFixed = useCallback((newAmountFixed: string) => {
    setInputState(state => ({
      ...state,
      amountQuoteInputFixed: newAmountFixed,
      amountLastChanged: 'quote',
    }));
  }, []);

  const setAmountBasePercentOfBalance = useCallback(
    (percentage: number) => {
      const bnBalance = new BigNumber(baseBalance.amount.toString()).times(percentage / 100);
      setAmountBaseFixed(bnBalance.toFixed(0));
    },
    [baseBalance.amount, setAmountBaseFixed]
  );

  const setAmountQuotePercentOfBalance = useCallback(
    (percentage: number) => {
      const bnBalance = new BigNumber(quoteBalance.amount.toString()).times(percentage / 100);
      setAmountQuoteFixed(bnBalance.toFixed(0));
    },
    [quoteBalance.amount, setAmountQuoteFixed]
  );

  const [{ amountB: calculatedOtherAmount, drops }, dropsStatus] = useAsyncPoll(
    useCallback(
      async () =>
        onex?.dropCoin({
          denomA: amountLastChanged === 'base' ? baseAsset.id : quoteAsset.id,
          denomB: amountLastChanged === 'base' ? quoteAsset.id : baseAsset.id,
          amountA: amountLastChanged === 'base' ? amountBaseInputFixed : amountQuoteInputFixed,
        }) ?? {
          amountB: '0',
          drops: '0',
        },
      [
        onex,
        amountBaseInputFixed,
        amountQuoteInputFixed,
        amountLastChanged,
        baseAsset.id,
        quoteAsset.id,
      ]
    ),
    {
      amountB: '0',
      drops: '0',
    }
  );

  const dropsToPoolRatio = useMemo(() => {
    const bnPoolDrops = new BigNumber(pool?.drops ?? 0);
    const bnDrops = new BigNumber(drops);
    const bnTotal = bnDrops.plus(bnPoolDrops);
    return bnDrops.dividedBy(bnTotal).toNumber();
  }, [pool?.drops, drops]);

  const amountBaseFixed =
    amountLastChanged === 'base' ? amountBaseInputFixed : calculatedOtherAmount;
  const amountQuoteFixed =
    amountLastChanged === 'quote' ? amountQuoteInputFixed : calculatedOtherAmount;

  const [amountQuote, amountQuoteStatus] = useDetailedBalance(quoteAsset, amountQuoteFixed);
  const [amountBase, amountBaseStatus] = useDetailedBalance(baseAsset, amountBaseFixed);

  const [amountBaseInput, setAmountBaseInput] = useFixedAmountInput({
    value: amountBaseFixed,
    onChangeValue: setAmountBaseFixed,
    meta: baseAsset.meta,
  });

  const [amountQuoteInput, setAmountQuoteInput] = useFixedAmountInput({
    value: amountQuoteFixed,
    onChangeValue: setAmountQuoteFixed,
    meta: baseAsset.meta,
  });

  const inputParams: Parameters<OnexClient['createDrop']>[0] = useMemo(() => {
    return {
      pair,
      drops,
    };
  }, [pair, drops]);

  const validationErrors = useMemo(() => {
    const errors: {
      drops?: string;
      amountBaseInput?: string;
      amountQuoteInput?: string;
    } = {};

    const bnDrops = new BigNumber(drops);
    const bnAmountBase = new BigNumber(amountBaseFixed);
    const bnAmountQuote = new BigNumber(amountQuoteFixed);

    if (bnDrops.isNaN() || bnDrops.lte(0)) {
      errors.drops = 'Drops must be a number > 0';
    }

    if (bnAmountBase.isNaN() || bnAmountBase.lte(0)) {
      errors.amountBaseInput = 'Must be a number > 0';
    }

    if (bnAmountQuote.isNaN() || bnAmountQuote.lte(0)) {
      errors.amountQuoteInput = 'Must be a number > 0';
    }

    if (bnAmountBase.gt(baseBalance.amount.toString())) {
      errors.amountBaseInput = 'Exceeds available balance';
    }

    if (bnAmountQuote.gt(quoteBalance.amount.toString())) {
      errors.amountQuoteInput = 'Exceeds available balance';
    }

    return Object.keys(errors).length > 0 ? errors : null;
  }, [drops, amountBaseFixed, amountQuoteFixed, baseBalance.amount, quoteBalance.amount]);

  const loadingStatus = useAsyncStatuses([balanceStatus]);
  const canSubmit = !validationErrors && !loadingStatus.pending && dropsStatus.finished;

  const submitTx = useCallback(
    (e?: React.FormEvent) => {
      e?.preventDefault();
      if (!canSubmit) return Promise.resolve();
      resetTxStatus();
      setTxResult(null);
      return attemptTx(async () => setTxResult((await onex?.createDrop(inputParams)) ?? null));
    },
    [onex, inputParams, canSubmit, resetTxStatus, attemptTx]
  );

  const reset = useCallback(() => {
    resetTxStatus();
    setTxResult(null);
  }, [resetTxStatus]);

  useEffect(reset, [baseAsset.id, quoteAsset.id, reset]);

  return {
    txStatus,
    txResult,
    loadingStatus,
    baseBalanceStatus,
    quoteBalanceStatus,
    balanceStatus,
    currentPricePerBaseStatus,
    currentPricePerQuoteStatus,
    amountBaseStatus,
    amountQuoteStatus,
    dropsStatus,

    baseAsset,
    quoteAsset,
    pool,

    drops,
    amountBaseInput,
    amountQuoteInput,

    currentPricePerBase,
    currentPricePerQuote,
    amountBase,
    amountQuote,
    baseBalance,
    quoteBalance,
    dropsToPoolRatio,

    validationErrors,
    canSubmit,

    setAmountBaseInput,
    setAmountQuoteInput,
    setAmountBasePercentOfBalance,
    setAmountQuotePercentOfBalance,

    submitTx,
    reset,
  };
}

export type AddDropsFormState = ReturnType<typeof useAddDropsForm>;
