import { useCallback, useMemo, useState } from 'react';

import { CosmosProposalType, OnexProposals } from '@onomy/client';
import { Blockchain } from '@onomy/frontend-relay-types';
import { useAddress } from '@onomy/react-hub';
import { useAsyncStatuses } from '@onomy/react-utils';
import { Metadata } from '@onomy/telescope/dist/codegen/cosmos/bank/v1beta1/bank';

// TODO: verify these limits or pull from chain?
const TITLE_MAX_LENGTH = 140;
const DESCRIPTION_MAX_LENGTH = 2500;
const BLANK_JSON = `{
  "name": "Cosmos Hub Atom",
  "symbol": "ATOM",
  "description": "The native staking token of the Cosmos Hub.",
  "denomUnits": [
    {
      "denom": "uatom",
      "exponent": 0,
      "aliases": [
        "microatom"
      ]
    },
    {
      "denom": "matom",
      "exponent": 3,
      "aliases": [
        "milliatom"
      ]
    },
    {
      "denom": "atom",
      "exponent": 6,
      "aliases": []
    }
  ],
  "base": "uatom",
  "display": "atom"
}`;

export function useDenomMetadataProposalContent({
  chain,
  type = OnexProposals.tree.pendulumlabs.market.market.DenomMetadataProposal,
}: {
  chain: Blockchain;
  type?: CosmosProposalType;
}) {
  const [address, addressStatus] = useAddress(chain);

  const [title, setTitle] = useState('');
  const [description, setDescription] = useState('');
  const [metadataJson, setMetadataJson] = useState(BLANK_JSON);

  const parsed = useMemo(() => {
    try {
      return JSON.parse(metadataJson);
    } catch (e) {
      return null;
    }
  }, [metadataJson]);

  const metadata = useMemo(() => {
    if (!parsed) return null;
    try {
      return Metadata.toProtoMsg(parsed);
    } catch (e) {
      return null;
    }
  }, [parsed]);

  const inputParams = useMemo(() => {
    return {
      sender: address?.id ?? '', // TODO
      title,
      description,
      metadata: metadata ? parsed : null,
    };
  }, [title, description, metadata, parsed, address?.id]);

  const loadingStatus = useAsyncStatuses([addressStatus]);

  const formErrors = useMemo(() => {
    const params = inputParams;

    const errors: {
      title?: string;
      description?: string;
      metadataJson?: string;
    } = {};

    if (!params.title) {
      errors.title = 'Title is required';
    } else if (params.title.length >= TITLE_MAX_LENGTH) {
      errors.title = `Title must be less than ${TITLE_MAX_LENGTH} characters`;
    }

    if (!params.description) {
      errors.description = 'Description is required';
    } else if (params.description.length >= DESCRIPTION_MAX_LENGTH) {
      errors.description = `Description must be less than ${DESCRIPTION_MAX_LENGTH} characters`;
    }

    if (parsed && !inputParams.metadata) {
      errors.metadataJson = 'Invalid format for metadata json';
    } else if (!parsed) {
      errors.metadataJson = 'Metadata is not valid JSON';
    }

    return Object.keys(errors).length > 0 ? errors : null;
  }, [inputParams, parsed]);

  const canSubmit = useMemo(() => {
    return !formErrors;
  }, [formErrors]);

  const reset = useCallback(() => {
    setTitle('');
    setDescription('');
    setMetadataJson(BLANK_JSON);
  }, []);

  const encoded = useMemo(() => {
    if (!canSubmit || !inputParams.metadata) return null;
    try {
      return (
        type.encode as typeof OnexProposals.tree.pendulumlabs.market.market.DenomMetadataProposal.encode
      )(inputParams);
    } catch (e) {
      // eslint-disable-next-line no-console
      console.warn(`Error encoding ${type.typeUrl}`, e);
      return null;
    }
  }, [type, canSubmit, inputParams]);

  return {
    TITLE_MAX_LENGTH,
    DESCRIPTION_MAX_LENGTH,

    loadingStatus,
    addressStatus,

    chain,
    type,

    title,
    description,
    metadataJson,
    metadata: parsed,

    formErrors,

    setTitle,
    setDescription,

    setMetadataJson,

    canSubmit,

    encoded,

    reset,
  };
}
