import React, { createContext, useContext, useMemo } from 'react';
import { DetailedFungibleAssetBalance, OnomyHubConstants } from '@onomy/hub';
import {
  serializeAsset,
  useAssetInBalances,
  useDelegations,
  useDetailedBalance,
  useDetailedBalances,
  useFiatTotal,
  useHoldings,
} from '@onomy/react-hub';
import { useAsyncStatuses } from '@onomy/react-utils';
import BigNumber from 'bignumber.js';
import { FungibleAssetOnChain } from 'index';
import { Blockchain } from '@onomy/frontend-relay-types';
import { PlaceholderErrorDisplay } from '../../components/PlaceholderErrorDisplay';
import { useOnexDayTradeFrame } from '../../exchange/TheGraph/useOnexDayTradeFrame';
import { useOnexBurned, useOnexMyDrops } from '../../exchange/hooks';
import { useOnomyApp } from './OnomyApp';
import { useVesting } from '../hooks/useVesting';

function useOnomyDashboardState() {
  const {
    onomyNomAsset,
    onexOnexAsset,
    onexNomAsset,
    onexEthAsset,
    onexUsdtAsset,
    loadingStatus: appLoadingStatus,
  } = useOnomyApp();
  const holdings = useHoldings();
  const [vesting, vestingStatus] = useVesting();
  const [myLiquidity, myLiquidityStatus] = useOnexMyDrops();
  const [delegations, delegationsStatus] = useDelegations();
  const oneNomFixed = useMemo(
    () => new BigNumber('1').shiftedBy(onomyNomAsset.meta.decimalPlaces).toFixed(0),
    [onomyNomAsset.meta.decimalPlaces]
  );
  const [oneNomCoin] = useDetailedBalance(onomyNomAsset, oneNomFixed);

  const { principalTotalFiat, rewardTotalFiat, liquidityTotalFiat } = useMemo(() => {
    return myLiquidity.reduce(
      (res, item) => {
        res.principalTotalFiat += item.fiatPrincipal;
        res.rewardTotalFiat += item.fiatReward;
        res.liquidityTotalFiat += item.fiatTotal;
        return res;
      },
      {
        principalTotalFiat: 0,
        rewardTotalFiat: 0,
        liquidityTotalFiat: 0,
      }
    );
  }, [myLiquidity]);

  const onexBalances = useMemo(
    () => holdings.balances.filter(x => x.asset.chain.name === OnomyHubConstants.ONEX_CHAIN_ID),
    [holdings.balances]
  );

  const mainBalances = useMemo(
    () => holdings.balances.filter(x => x.asset.chain.name !== OnomyHubConstants.ONEX_CHAIN_ID),
    [holdings.balances]
  );

  const onexBalancesUnlocked = useMemo(
    () =>
      holdings.balancesUnlocked.filter(x => x.asset.chain.name === OnomyHubConstants.ONEX_CHAIN_ID),
    [holdings.balancesUnlocked]
  );

  const mainBalancesUnlocked = useMemo(
    () =>
      holdings.balancesUnlocked.filter(x => x.asset.chain.name !== OnomyHubConstants.ONEX_CHAIN_ID),
    [holdings.balancesUnlocked]
  );

  const onexBalancesLocked = useMemo(
    () =>
      holdings.balancesLocked.filter(x => x.asset.chain.name === OnomyHubConstants.ONEX_CHAIN_ID),
    [holdings.balancesLocked]
  );

  const mainBalancesLocked = useMemo(
    () =>
      holdings.balancesLocked.filter(x => x.asset.chain.name !== OnomyHubConstants.ONEX_CHAIN_ID),
    [holdings.balancesLocked]
  );

  const onexBalancesTotalFiat = useFiatTotal(onexBalances);
  const mainBalancesTotalFiat = useFiatTotal(mainBalances);
  const onexBalancesUnlockedTotalFiat = useFiatTotal(onexBalancesUnlocked);
  const mainBalancesUnlockedTotalFiat = useFiatTotal(mainBalancesUnlocked);
  const onexBalancesLockedTotalFiat = useFiatTotal(onexBalancesLocked);
  const mainBalancesLockedTotalFiat = useFiatTotal(mainBalancesLocked);

  const balancesUnlockedTotalFiat = onexBalancesUnlockedTotalFiat + mainBalancesUnlockedTotalFiat;
  const balancesLockedTotalFiat = onexBalancesLockedTotalFiat + mainBalancesLockedTotalFiat;

  const nomStaked = useAssetInBalances(onomyNomAsset, holdings.stakedBalances);
  const nomRewards = useAssetInBalances(onomyNomAsset, holdings.rewardBalances);
  const nomUnbonding = useAssetInBalances(onomyNomAsset, holdings.unbondingBalances);

  const mainGrandTotalFiat =
    nomStaked.fiatValue + nomRewards.fiatValue + nomUnbonding.fiatValue + mainBalancesTotalFiat;

  const [nomUsdtDay, nomUsdtDayStatus] = useOnexDayTradeFrame({
    baseAsset: onexNomAsset,
    quoteAsset: onexUsdtAsset,
  });

  const [nomEthDay, nomUsdtEthStatus] = useOnexDayTradeFrame({
    baseAsset: onexNomAsset,
    quoteAsset: onexEthAsset,
  });

  const [onexNomBurned, onexNomBurnedStatus] = useOnexBurned();

  const onexStaked = useAssetInBalances(onexOnexAsset, holdings.stakedBalances);
  const onexRewards = useAssetInBalances(onexOnexAsset, holdings.rewardBalances);
  const onexUnbonding = useAssetInBalances(onexOnexAsset, holdings.unbondingBalances);

  const onexGrandTotalFiat =
    onexBalancesTotalFiat +
    liquidityTotalFiat +
    onexStaked.fiatValue +
    onexRewards.fiatValue +
    onexUnbonding.fiatValue;
  const grandTotalFiat = mainGrandTotalFiat + onexGrandTotalFiat;

  const loadingStatus = useAsyncStatuses([
    appLoadingStatus,
    holdings.loadingStatus,
    vestingStatus,
    nomUsdtDayStatus,
    nomUsdtEthStatus,
    myLiquidityStatus,
    delegationsStatus,

    onexNomBurnedStatus,
  ]);

  return {
    holdings,
    vesting,
    myLiquidity,
    delegations,

    nomUsdtDay,
    nomEthDay,
    onexNomBurned,

    onexBalances,
    onexBalancesUnlocked,
    onexBalancesLocked,
    mainBalances,
    mainBalancesUnlocked,
    mainBalancesLocked,

    oneNomCoin,

    onexStaked,
    onexRewards,
    onexUnbonding,

    principalTotalFiat,
    rewardTotalFiat,
    liquidityTotalFiat,

    onexBalancesTotalFiat,
    onexBalancesUnlockedTotalFiat,
    onexBalancesLockedTotalFiat,
    mainBalancesTotalFiat,
    mainBalancesUnlockedTotalFiat,
    mainBalancesLockedTotalFiat,
    balancesUnlockedTotalFiat,
    balancesLockedTotalFiat,
    mainGrandTotalFiat,
    onexGrandTotalFiat,
    grandTotalFiat,

    loadingStatus,
  };
}

export type OnomyDashboardContextType = ReturnType<typeof useOnomyDashboardState>;

const OnomyDashboardContext = createContext<OnomyDashboardContextType>(null as any);

export function useOnomyDashboard() {
  const state = useContext(OnomyDashboardContext);
  if (!state) throw new Error('OnomyDashboardContext missing');
  return state;
}

// eslint-disable-next-line @typescript-eslint/naming-convention
function DEFAULT_ERROR_RENDER_FN(error: any) {
  return <PlaceholderErrorDisplay error={error} />;
}

export function OnomyDashboardProvider({
  children,
  renderLoading,
  renderError = DEFAULT_ERROR_RENDER_FN,
}: {
  children: JSX.Element | JSX.Element[];
  renderLoading?: () => JSX.Element | JSX.Element[] | null;
  renderError?: typeof DEFAULT_ERROR_RENDER_FN;
}) {
  const state = useOnomyDashboardState();

  if (state.loadingStatus.pending && renderLoading) return renderLoading();
  if (state.loadingStatus.error) return renderError(state.loadingStatus.error);
  return <OnomyDashboardContext.Provider value={state}>{children}</OnomyDashboardContext.Provider>;
}

function useFindBalance(assetId: string, balances: DetailedFungibleAssetBalance[]) {
  return React.useMemo(
    () =>
      balances.find(a => serializeAsset(a.asset) === assetId) ?? OnomyHubConstants.UNKNOWN_BALANCE,
    [balances, assetId]
  );
}

export function useAvailableBalance(idOrAsset: string | FungibleAssetOnChain) {
  const { holdings } = useOnomyDashboard();
  const id = typeof idOrAsset === 'string' ? idOrAsset : serializeAsset(idOrAsset);
  return useFindBalance(id, holdings.balances);
}

export function useStakedBalance(idOrAsset: string | FungibleAssetOnChain) {
  const { holdings } = useOnomyDashboard();
  const id = typeof idOrAsset === 'string' ? idOrAsset : serializeAsset(idOrAsset);
  return useFindBalance(id, holdings.stakedBalances);
}

export function useRewardBalance(idOrAsset: string | FungibleAssetOnChain) {
  const { holdings } = useOnomyDashboard();
  const id = typeof idOrAsset === 'string' ? idOrAsset : serializeAsset(idOrAsset);
  return useFindBalance(id, holdings.rewardBalances);
}

export function useUnstakingBalance(idOrAsset: string | FungibleAssetOnChain) {
  const { holdings } = useOnomyDashboard();
  const id = typeof idOrAsset === 'string' ? idOrAsset : serializeAsset(idOrAsset);
  return useFindBalance(id, holdings.unbondingBalances);
}

export function useAssetDelegations(idOrAsset: string | FungibleAssetOnChain) {
  const { delegations } = useOnomyDashboard();
  const id = typeof idOrAsset === 'string' ? idOrAsset : serializeAsset(idOrAsset);

  return React.useMemo(
    () => delegations.filter(a => serializeAsset(a.delegated.asset) === id),
    [delegations, id]
  );
}

export function useChainDelegations(idOrChain: string | Blockchain) {
  const { delegations } = useOnomyDashboard();
  const id = typeof idOrChain === 'string' ? idOrChain : idOrChain.name;

  return React.useMemo(
    () => delegations.filter(a => a.delegated.asset.chain.name === id),
    [delegations, id]
  );
}
